"""
Contains the core encryption logic
"""

from Crypto.Cipher import AES
import base64
from pbkdf2 import PBKDF2
from .conf import ENCRYPTION_KEY


class EncryptionClient:

    def __init__(self):
        self.iv = bytes([83, 71, 26, 58, 54, 35, 22, 11,
                         83, 71, 26, 58, 54, 35, 22, 11])
        self.enc_key = ENCRYPTION_KEY

    def _pad_and_convert_to_bytes(self, text):
        return bytes(text+chr(16-len(text) % 16)*(16-len(text) % 16), encoding="utf-8")

    def encrypt(self, text):
        padded_text = self._pad_and_convert_to_bytes(text)
        key_gen = PBKDF2(self.enc_key, self.iv)

        aesiv = key_gen.read(16)
        aeskey = key_gen.read(32)
        cipher = AES.new(aeskey, AES.MODE_CBC, aesiv)

        return str(base64.b64encode(cipher.encrypt(padded_text)), encoding="utf-8")



"""
Contains reusable payloads across requests
"""
from .conf import APP_NAME, USER_ID, PASSWORD, USER_KEY

HEADERS = {'Content-Type': 'application/json'}

GENERIC_PAYLOAD = {
    "head": {
        "appName": APP_NAME,
        "appVer": "1.0",
        "key": USER_KEY,
        "osName": "WEB",
        "requestCode": "",
        "userId": USER_ID,
        "password": PASSWORD
    },
    "body": {
        "ClientCode": ""
    }
}

LOGIN_PAYLOAD = {"head": {
    "appName": APP_NAME,
    "appVer": "1.0",
    "key": USER_KEY,
    "osName": "WEB",
    "requestCode": "5PLoginV2",
    "userId": USER_ID,
    "password": PASSWORD
},
    "body":
    {
    "Email_id": "",
    "Password": "",
    "LocalIP": "",
    "PublicIP": "",
    "HDSerailNumber": "",
    "MACAddress": "",
    "MachineID": "039377",
    "VersionNo": "1.7",
    "RequestNo": "1",
    "My2PIN": "",
    "ConnectionType": "1"
}
}



import requests
from .auth import EncryptionClient
from .const import GENERIC_PAYLOAD, HEADERS
from .conf import APP_SOURCE
from .order import *
import time


class FivePaisaClient:

    LOGIN_ROUTE = "https://Openapi.5paisa.com/VendorsAPI/Service1.svc/V2/LoginRequestMobileNewbyEmail"

    MARGIN_ROUTE = "https://Openapi.5paisa.com/VendorsAPI/Service1.svc/V3/Margin"
    ORDER_BOOK_ROUTE = "https://Openapi.5paisa.com/VendorsAPI/Service1.svc/V2/OrderBook"
    HOLDINGS_ROUTE = "https://openapi.5paisa.com/VendorsAPI/Service1.svc/V2/Holding"
    POSITIONS_ROUTE = "https://Openapi.5paisa.com/VendorsAPI/Service1.svc/V1/NetPositionNetWise"

    ORDER_PLACEMENT_ROUTE = "https://Openapi.5paisa.com/VendorsAPI/Service1.svc/V1/OrderRequest"
    ORDER_STATUS_ROUTE = "https://Openapi.5paisa.com/VendorsAPI/Service1.svc/OrderStatus"
    TRADE_INFO_ROUTE = "https://Openapi.5paisa.com/VendorsAPI/Service1.svc/TradeInformation"

    MARGIN_REQUEST_CODE = "5PMarginV3"
    ORDER_BOOK_REQUEST_CODE = "5POrdBkV2"
    HOLDINGS_REQUEST_CODE = "5PHoldingV2"
    POSITIONS_REQUEST_CODE = "5PNPNWV1"

    def __init__(self, email=None, passwd=None, dob=None):
        """
        Main constructor for client.
        Expects user's email, password and date of birth in YYYYMMDD format.
        """
        self.email = email
        self.passwd = passwd
        self.dob = dob
        self.payload = GENERIC_PAYLOAD
        self.client_code = None
        self.session = requests.Session()

    def login(self):
        encryption_client = EncryptionClient()
        secret_email = encryption_client.encrypt(self.email)
        secret_passwd = encryption_client.encrypt(self.passwd)
        secret_dob = encryption_client.encrypt(self.dob)
        self.payload["body"]["Email_id"] = secret_email
        self.payload["body"]["Password"] = secret_passwd
        self.payload["body"]["My2PIN"] = secret_dob
        self.payload["head"]["requestCode"] = "5PLoginV2"
        res = self._login_request(self.LOGIN_ROUTE)
        self._set_client_code(res["body"]["ClientCode"])
        return self.client_code

    def holdings(self):
        return self._user_info_request("HOLDINGS")

    def margin(self):
        return self._user_info_request("MARGIN")

    def order_book(self):
        return self._user_info_request("ORDER_BOOK")

    def positions(self):
        return self._user_info_request("POSITIONS")

    def _login_request(self, route):
        res = self.session.post(route, json=self.payload, headers=HEADERS)
        return res.json()

    def _set_client_code(self, client_code):
        self.client_code = client_code

    def _user_info_request(self, data_type):
        payload = GENERIC_PAYLOAD
        payload["body"]["ClientCode"] = self.client_code
        if data_type == "MARGIN":
            request_code = self.MARGIN_REQUEST_CODE
            url = self.MARGIN_ROUTE
        elif data_type == "ORDER_BOOK":
            request_code = self.ORDER_BOOK_REQUEST_CODE
            url = self.ORDER_BOOK_ROUTE
        elif data_type == "HOLDINGS":
            request_code = self.HOLDINGS_REQUEST_CODE
            url = self.HOLDINGS_ROUTE
        elif data_type == "POSITIONS":
            request_code = self.POSITIONS_REQUEST_CODE
            url = self.POSITIONS_ROUTE
        else:
            raise Exception("Invalid data type requested")

        payload["head"]["requestCode"] = request_code
        response = self.session.post(url, json=payload, headers=HEADERS)
        return response.json()

    def order_request(self, req_type) -> dict:

        self.payload["body"]["ClientCode"] = self.client_code

        if req_type == "OP":
            url = self.ORDER_PLACEMENT_ROUTE
            self.payload["head"]["requestCode"] = "5POrdReq"
        elif req_type == "OS":
            url = self.ORDER_STATUS_ROUTE
            self.payload["head"]["requestCode"] = "5POrdStatus"
        elif req_type == "TI":
            url = self.TRADE_INFO_ROUTE
            self.payload["head"]["requestCode"] = "5PTrdInfo"
        else:
            raise Exception("Invalid request type!")

        res = self.session.post(url, json=self.payload,
                                headers=HEADERS)
        return res.json()

    def fetch_order_status(self, req_list: RequestList) -> dict:
        self.payload["body"]["OrdStatusReqList"] = req_list.orders
        return self.order_request("OS")

    def fetch_trade_info(self, req_list: RequestList) -> dict:
        self.payload["body"]["TradeDetailList"] = req_list.orders
        return self.order_request("TI")

    def place_order(self, order: Order) -> dict:

        self.payload["body"]["OrderFor"] = order.order_for
        self.payload["body"]["Exchange"] = order.exchange
        self.payload["body"]["ExchangeType"] = order.exchange_type
        self.payload["body"]["Price"] = order.price
        self.payload["body"]["OrderID"] = order.order_id
        self.payload["body"]["OrderType"] = order.order_type
        self.payload["body"]["Qty"] = order.quantity
        # Probably UNIX timestamp
        self.payload["body"]["OrderDateTime"] = f"/Date({int(time.time())})/"
        self.payload["body"]["ScripCode"] = order.scrip_code
        self.payload["body"]["AtMarket"] = str(order.atmarket).lower()
        self.payload["body"]["RemoteOrderID"] = order.remote_order_id
        self.payload["body"]["ExchOrderID"] = order.exch_order_id
        self.payload["body"]["DisQty"] = order.disqty
        self.payload["body"]["IsStopLossOrder"] = str(
            order.is_stoploss_order).lower()
        self.payload["body"]["StopLossPrice"] = order.stoploss_price
        self.payload["body"]["IsVTD"] = str(order.is_vtd).lower()
        self.payload["body"]["IOCOrder"] = str(order.ioc_order).lower()
        self.payload["body"]["IsIntraday"] = str(order.is_intraday).lower()
        self.payload["body"]["PublicIP"] = order.public_ip
        self.payload["body"]["AHPlaced"] = order.ahplaced
        # Hardcoded for now
        self.payload["body"]["ValidTillDate"] = f"/Date({int(time.time())})/"
        self.payload["body"]["TradedQty"] = order.traded_qty
        self.payload["body"]["OrderRequesterCode"] = self.client_code
        self.payload["body"]["AppSource"] = APP_SOURCE
        self.payload["body"]["iOrderValidity"] = order.order_validity
        return self.order_request("OP")



"""
Get your API keys from https://www.5paisa.com/developerapi/apikeys
"""
import configparser

config = configparser.ConfigParser()
config.read("keys.conf")
try:
    section = config["KEYS"]
except KeyError:
    raise Exception("Please configure your keys in keys.conf")

APP_SOURCE = config["KEYS"]["APP_SOURCE"]
APP_NAME = config["KEYS"]["APP_NAME"]
USER_ID = config["KEYS"]["USER_ID"]
PASSWORD = config["KEYS"]["PASSWORD"]
USER_KEY = config["KEYS"]["USER_KEY"]
ENCRYPTION_KEY = config["KEYS"]["ENCRYPTION_KEY"]



"""
Contains base classes for Orders etc.
"""
from .const import GENERIC_PAYLOAD, HEADERS
import requests
from .conf import APP_SOURCE


class Exchange:

    NSE = "N"
    BSE = "B"
    MCX = "M"


class ExchangeType:

    CASH = "C"
    DERIVATIVE = "D"
    CURRENCY = "U"


class OrderType:

    PLACE = "P"
    MODIFY = "M"
    CANCEL = "C"


class Order:

    def __init__(self, order_for: str, exchange: str, exchange_type: str, price: float,
                 order_id: int, order_type: str, quantity: int,
                 scrip_code: int, atmarket: bool, remote_order_id: int, exch_order_id: int, disqty: int,
                 stoploss_price: float, is_stoploss_order: bool, ioc_order: bool, is_intraday: bool, is_vtd: bool, vtd: str,
                 ahplaced: str, public_ip: str, order_validity: int, traded_qty: int):

        self.order_for = order_for
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.price = price
        self.order_id = order_id
        self.order_type = order_type
        self.quantity = quantity
        self.scrip_code = scrip_code
        self.atmarket = atmarket
        self.remote_order_id = remote_order_id
        self.exch_order_id = exch_order_id
        self.disqty = disqty
        self.stoploss_price = stoploss_price
        self.is_stoploss_order = is_stoploss_order
        self.ioc_order = ioc_order
        self.is_intraday = is_intraday
        self.is_vtd = is_vtd
        self.vtd = vtd
        self.ahplaced = ahplaced
        self.public_ip = public_ip
        self.order_validity = order_validity
        self.traded_qty = traded_qty
        self.app_source = int(APP_SOURCE)


class OrderForStatus:
    """
    Order class for representing an order for only Status and trade information request.
    """

    def __init__(self, exchange: str, exchange_type: str, scrip_code: int, order_id: str) -> None:
        self.exchange = exchange
        self.exchange_type = exchange_type
        self.scrip_code = scrip_code
        self.order_id = order_id

    def __str__(self):
        return f"Order -> Exchange:{self.exchange}, Type: {self.exchange_type}, Scrip Code: {self.scrip_code}, Order ID: {self.order_id}"

    def __repr__(self) -> dict:
        """
        Overriding repr to return dict which can be directly appended to the orders list.
        """
        return {
            "Exch": self.exchange,
            "ExchType": self.exchange_type,
            "ScripCode": self.scrip_code,
            "RemoteOrderID": self.order_id
        }


class RequestList:

    def __init__(self):
        self.orders = []

    def add_order(self, order: OrderForStatus) -> None:
        """
        Adds order to the RequestList object required in the request payload.
        """
        self.orders.append(order.__repr__())
